import {
  ref,
  unref,
  computed,
  watchEffect,
  getCurrentInstance,
} from 'vue-demi';
import { add, subtract, bignumber, format } from 'mathjs';
import { isDef, isFunction } from '../libs/helper';
const isNumber = n => isDef(n) && !isNaN(unref(n));
function useCounter(initialValue, options = {}) {
  if (!getCurrentInstance()) {
    throw new Error(
      'Invalid hook call: `useCounter` can only be called inside of `setup()`.',
    );
  }
  const { min, max, step } = options;
  const initial = () =>
    isNumber(initialValue) ? Number(unref(initialValue)) : 0;
  const $step = () => (isNumber(step) ? Number(unref(step)) : 1);
  const fix = num => {
    let result = +unref(num);
    if (isNumber(max)) {
      result = Math.min(Number(unref(max)), result);
    }
    if (isNumber(min)) {
      result = Math.max(Number(unref(min)), result);
    }
    return result;
  };
  const current = ref(fix(initial()));
  const set = v => {
    current.value = fix(isFunction(v) ? v(current.value) : v);
  };
  const inc = v =>
    set(
      Number(
        format(
          add(
            bignumber(current.value),
            bignumber(isDef(v) && isNumber(v) ? unref(v) : $step()),
          ),
        ),
      ),
    );
  const dec = v =>
    set(
      Number(
        format(
          subtract(
            bignumber(current.value),
            bignumber(isDef(v) && isNumber(v) ? unref(v) : $step()),
          ),
        ),
      ),
    );
  const reset = () => set(fix(initial()));
  const count = computed({
    get: () => current.value,
    set: v => {
      if (isNumber(v)) {
        set(v);
      } else {
        throw new TypeError(
          `Invalid assignment: expected a number-string or number but got: ${
            isDef(v) ? typeof v : v
          }`,
        );
      }
    },
  });
  watchEffect(
    () => {
      // if `min`/`max` is RefTyped, collect dependencies:
      unref(min);
      unref(max);
      set(current.value);
    },
    {
      flush: 'sync',
    },
  );
  return {
    count,
    inc,
    dec,
    set,
    reset,
  };
}
export default useCounter;
